import pytest

from typing import Annotated
from types import ModuleType

# from unittest.mock import MagicMock, patch

from kaya_module_sdk.src.module.template import Module, Config
from kaya_module_sdk.src.module.arguments import Args
from kaya_module_sdk.src.module.returns import Rets
from kaya_module_sdk.src.utils.generators.methods import kaya_io
from kaya_module_sdk.src.utils.metadata.abstract_metadata import KMetadata
from kaya_module_sdk.src.utils.metadata.abstract_validation import KValidation

# DUMMY DATA


class DummyResource:
    """Dummy resource classes to simulate importlib.resources behavior."""

    def __init__(self, file_path):
        self.dummy_path = file_path

    def joinpath(self, file_name):
        # For testing we ignore file_name and simply return self.
        return self


class DummyContextManager:
    def __init__(self, file_path):
        self.file_path = file_path

    def __enter__(self):
        return self.file_path

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass


class ArgsType(Args):
    key1: str
    key2: str

    def set_key1(self, value):
        pass

    def set_key2(self, value):
        pass


class RetsType(Rets):
    key1: str
    key2: str

    def __init__(self, *args, **kwargs):
        pass


@kaya_io()
class DecoratedArgsType(Args):
    _field1: Annotated[int, "metadata"]
    _field2: Annotated[str, "metadata"]


class DummyConfig(Config):
    def __init__(self, name="dummy", version="1.0"):
        self.name = name
        self.version = version
        self._mandatory = []  # NOTE: For simplicity, no mandatory fields


class TestModule(Module):
    def __init__(self, manifest=None, config=None):
        if manifest:
            self.manifest = manifest
        #       if config:
        self.config = config or DummyConfig()
        # mod1 has a valid manifest; mod2 is missing a manifest.
        self.modules = {
            "mod1": DummyModule(
                manifest={
                    "inputs": [{"label": "a", "validations": ["rule1"]}],
                    "outputs": [],
                },
                config=DummyConfig(),
            ),
            "mod2": DummyModule(manifest=None, config=DummyConfig()),
        }

    def __repr__(self):
        return "DummyModule"

    def main(self, args) -> Rets:
        return Rets()  # NOTE: Simulating main method behavior for testing


class DummyModule:
    def __init__(self, manifest, config=None):
        self.manifest = manifest
        self.config = config or DummyConfig()

    def __repr__(self):
        return "DummyModule"


class KMetadataSubclass(KMetadata):
    def __init__(self, data):
        self._data = data

    def get_data(self):
        return self._data

    def __str__(self) -> str:
        return f"KMetadata: {self._data}"


class KValidationSubclass(KValidation):
    def __init__(self, data):
        self._data = data

    def get_data(self):
        return self._data

    def __str__(self) -> str:
        return f"KValidation: {self._data}"


class DummyKVLE:
    """Dummy implementations for KVLE to isolate KVL tests."""

    def __init__(self, **context):
        self.context = context

    def check_rules(self):
        # Return a dummy rules result
        return {"dummy_rule": "passed"}

    def check_meta(self, module_package):
        # Return a dummy metadata result
        return {"dummy_meta": "passed"}

    def check_source(self, python_files):
        # Return a dummy source result
        return {"dummy_source": "passed"}


class DummyKVLR:
    """Dummy implementations for KVLR to isolate KVL tests."""

    def __init__(self, **context):
        self.context = context

    def generate_report(self, data, dump=False):
        return {"report": data}


class DummyKVLEForSuccess:
    def check_source(self, python_files):
        # Return a dummy "success" result.
        # Format: a dict with keys "ok" and "nok".
        file_key = list(python_files.keys())[0] if python_files else "unknown"
        return {"ok": [{"path": file_key, "errors": [], "result": "OK"}], "nok": []}


class DummyKVLRForSuccess:
    def generate_report(self, data, dump=False):
        # Simply wrap the provided data in a dict under "report".
        return {"report": data}


class DummyKVLEWithSourceError:
    def check_source(self, python_files):
        # Return a dummy error structure:
        file_key = list(python_files.keys())[0] if python_files else "unknown"
        return {
            "ok": [],
            "nok": [{"path": file_key, "errors": ["dummy error"], "result": "NOK"}],
        }


class DummyKVLRForErrors:
    def generate_report(self, data, dump=False):
        return {"report": data}


class DummyConstraint:
    """Dummy constraint class for testing load_constraint_rules"""

    def __init__(self, a, b=None):
        self._data = {"dummy": "value"}

    def load(self, rule):
        # NOTE: For testing purposes, we ignore the rule and assume successful load.
        pass


class DummyModuleInstance:
    def __init__(self):
        self.config = DummyConfig()
        # NOTE: mod1 has a valid manifest; mod2 is missing a manifest.
        self.modules = {
            "mod1": DummyModule(
                manifest={
                    "inputs": [{"label": "a", "validations": ["rule1"]}],
                    "outputs": [],
                },
                config=DummyConfig(),
            ),
            "mod2": DummyModule(manifest=None, config=DummyConfig()),
        }


# UTILS


def fake_import_module(name):
    """Fake import_module for module package loading."""
    # NOTE: When KVL.load_module_package calls import_module,
    #   we expect a module with a "KayaStrategyModule" attribute.
    if name == "dummy_module.module":
        dummy_mod = ModuleType("dummy_module.module")

        class DummyStrategyModule:
            def __init__(self):
                self.config = type("Config", (), {"name": "dummy", "version": "1.0"})

        setattr(dummy_mod, "KayaStrategyModule", DummyStrategyModule)
        return dummy_mod
    raise ModuleNotFoundError("Module not found: " + name)


def patch_metadata_classes_kvle(monkeypatch):
    """
    Patch all metadata constraint class names used in KVLE.load_constraint_rules
    to our DummyConstraint.
    """
    # NOTE: Import the module where KVLE is defined.
    import kaya_module_sdk.src.testing.kvl_executer as kvle_mod

    names = [
        "EQ",
        "GT",
        "GTE",
        "LT",
        "LTE",
        "MaxLen",
        "Max",
        "MinLen",
        "Min",
        "ValueRange",
        "EQLen",
        "Const",
        "NotConst",
        "Order",
    ]
    for name in names:
        monkeypatch.setattr(kvle_mod, name, DummyConstraint)


@pytest.fixture
def dummy_resource():
    return DummyResource


@pytest.fixture
def dummy_context_manager():
    return DummyContextManager


@pytest.fixture
def dummy_kvle():
    return DummyKVLE


@pytest.fixture
def dummy_kvlr():
    return DummyKVLR


@pytest.fixture
def dummy_kvle_ok():
    return DummyKVLEForSuccess


@pytest.fixture
def dummy_kvlr_ok():
    return DummyKVLRForSuccess


@pytest.fixture
def dummy_kvle_source_nok():
    return DummyKVLEWithSourceError


@pytest.fixture
def dummy_kvlr_nok():
    return DummyKVLRForErrors


@pytest.fixture
def dummy_config():
    return DummyConfig


@pytest.fixture
def dummy_constraint():
    return DummyConstraint


@pytest.fixture
def patch_meta_kvle():
    return patch_metadata_classes_kvle


@pytest.fixture
def import_module():
    return fake_import_module


@pytest.fixture
def mock_module():
    return TestModule()


@pytest.fixture
def mock_module_instance():
    return DummyModuleInstance()


@pytest.fixture
def kaya_io_decorated_args_type(request):
    instance = DecoratedArgsType()
    instance.set_field1(request.param.get("_field1"))
    instance.set_field2(request.param.get("_field2"))
    return instance


@pytest.fixture
def args_type():
    return ArgsType()


@pytest.fixture
def rets_type_obj():
    return RetsType()


@pytest.fixture
def rets_type_cls():
    return RetsType


@pytest.fixture
def kmetadata_subclass(request):
    instance = KMetadataSubclass(data=request.param.get("_data"))
    return instance


@pytest.fixture
def kvalidation_subclass(request):
    instance = KValidationSubclass(data=request.param.get("_data"))
    return instance


# CODE DUMP



from kaya_module_sdk.src.module.config import KConfig


def test_initialization_with_default_values():
    """Test Initialization with default values"""
    config = KConfig(name="Test", version="1.0", category="Category", author="Author")
    # NOTE: Assert default initialization
    assert config.name == "Test"
    assert config.version == "1.0"
    assert config.category == "Category"
    assert config.author == "Author"
    assert config.display_label == ""
    assert config.description == ""
    assert config.author_email == ""
    assert config.MANIFEST == {
        "PACKAGE": {
            "NAME": "Test",
            "LABEL": "",
            "VERSION": "1.0",
            "DESCRIPTION": "",
            "CATEGORY": "Category",
        },
        "MODULES": {},
    }
    assert config._mandatory == ["name", "version", "category", "author"]


def test_metadata():
    """Test metadata"""
    config = KConfig(name="Test", version="1.0", category="Category", author="Author")
    metadata = config.metadata()
    # NOTE: Check that metadata contains type hints for each attribute
    assert "name" in metadata
    assert "version" in metadata
    assert "display_label" in metadata
    assert "category" in metadata
    assert "description" in metadata
    assert "author" in metadata
    assert "author_email" in metadata
    assert "MANIFEST" in metadata
    assert "DEFAULTS" in metadata
    assert "_mandatory" in metadata


def test_format_metadata_with_no_modules():
    """Test the _format_metadata method"""
    config = KConfig(name="Test", version="1.0", category="Category", author="Author")
    result = config._format_metadata()
    # NOTE: Check if 'PACKAGE' keys are populated correctly
    assert "PACKAGE" in result
    assert result["PACKAGE"]["NAME"] == "Test"
    assert result["PACKAGE"]["VERSION"] == "1.0"
    assert result["PACKAGE"]["CATEGORY"] == "Category"
    # NOTE: Check MODULES is empty
    assert "MODULES" in result
    assert result["MODULES"] == {}


def test_initialization_with_kwargs():
    """Test Initialization with custom values"""
    config = KConfig(
        name="App",
        version="2.0",
        display_label="App Label",
        category="Utilities",
        author="Dev",
        author_email="dev@example.com",
    )
    # NOTE: Assert all values are correctly set
    assert config.name == "App"
    assert config.version == "2.0"
    assert config.display_label == "App Label"
    assert config.category == "Utilities"
    assert config.author == "Dev"
    assert config.author_email == "dev@example.com"
    assert config.MANIFEST == {
        "PACKAGE": {
            "NAME": "App",
            "LABEL": "App Label",
            "VERSION": "2.0",
            "DESCRIPTION": "",
            "CATEGORY": "Utilities",
        },
        "MODULES": {},
    }


def test_format_metadata_with_modules():
    # NOTE: Create a KConfig object for module1
    module = (
        "module1",
        KConfig(name="Module1", version="1.0", category="Category", author="Author"),
    )
    module[1].manifest = {}
    # NOTE: Create a KConfig object for the main config
    config = KConfig(name="Test", version="1.0", category="Category", author="Author")
    # NOTE: Call _format_metadata with the module
    result = config._format_metadata(module)
    # NOTE: Verify the structure of the result
    assert "MODULES" in result, "MODULES should be in the result."
    assert "module1" in result["MODULES"], "Module 'module1' should be in the MODULES."


def test_recompute_package_metadata():
    """Test recompute_package_metadata"""
    config = KConfig(name="Test", version="1.0", category="Category", author="Author")
    new_module = (
        "module2",
        KConfig(name="Module2", version="1.1", category="Category", author="Author"),
    )
    new_module[1].manifest = {}
    # NOTE: Recompute metadata and check if MANIFEST is updated
    result = config.recompute_package_metadata(new_module)
    assert "MODULES" in result
    assert "module2" in result["MODULES"]


def test_data_method():
    """Test data method"""
    config = KConfig(name="Test", version="1.0", category="Category", author="Author")
    # NOTE: Check if data() returns the expected dict with instance attributes
    data = config.data()
    assert data["name"] == "Test"
    assert data["version"] == "1.0"
    assert data["category"] == "Category"
    assert data["author"] == "Author"
    assert "MANIFEST" in data
    assert "_mandatory" in data



from kaya_module_sdk.src.utils.metadata.abstract_metadata import KMetadata
from kaya_module_sdk.src.utils.metadata.abstract_validation import KValidation


class Max(KMetadata, KValidation):
    def __init__(self, value: float | str) -> None:
        self._data = {
            "maximum": value,
        }

    def __str__(self) -> str:
        return f'max:{self._data["maximum"]}'

    def load(self, str_repr: str) -> dict:
        segmented = str_repr.split(":")
        if not len(segmented) == 2 or segmented[0] != "max":
            return {}
        try:
            self._data["maximum"] = float(segmented[1])
        except Exception:
            self._data["maximum"] = segmented[1]
        return self._data



import pytest

from kaya_module_sdk.src.utils.constraints.equal import keq

# UTILS


@keq(10)
def add(a, b):
    return a + b


@keq(5)
def multiply(a, b):
    return a * b


# TESTS


def test_valid_arguments():
    """Test cases where all arguments meet the condition."""
    assert add(10, 10) == 20
    assert multiply(5, 5) == 25


def test_invalid_arguments():
    """Test cases where an argument fails the condition."""
    with pytest.raises(ValueError, match="Value should be equal to 10"):
        add(10, 5)
    with pytest.raises(ValueError, match="Value should be equal to 5"):
        multiply(5, 10)


def test_mixed_types():
    """Test cases where non-numeric arguments are ignored."""

    @keq(100)
    def concat_numbers_and_strings(a, b):
        return f"{a}-{b}"

    # NOTE: Non-numeric arguments should not raise errors
    assert concat_numbers_and_strings("hello", "world") == "hello-world"
    assert concat_numbers_and_strings(100, "world") == "100-world"
    # NOTE: Numeric arguments that fail the condition should raise an error
    with pytest.raises(ValueError, match="Value should be equal to 100"):
        concat_numbers_and_strings(50, "world")


def test_no_arguments():
    """Test cases where the function receives no arguments."""

    @keq(42)
    def no_arg_function():
        return "no arguments"

    assert no_arg_function() == "no arguments"


def test_keyword_arguments():
    """
    Test cases where the function receives keyword arguments.
    The decorator only checks positional arguments.
    """

    @keq(7)
    def custom_func(a, b, c=None):
        return a + b

    # NOTE: Only positional arguments are checked
    assert custom_func(7, 7, c=14) == 14
    # NOTE: Invalid positional argument
    with pytest.raises(ValueError, match="Value should be equal to 7"):
        custom_func(7, 8, c=14)


def test_nested_functions():
    """Test cases where the decorated function is nested within another function."""

    def outer_function():
        @keq(3)
        def inner_function(a):
            return a * 2

        return inner_function

    inner = outer_function()
    assert inner(3) == 6  # Valid input
    with pytest.raises(ValueError, match="Value should be equal to 3"):
        inner(4)


def test_edge_cases():
    """Test cases for edge values like zero and negative numbers."""

    @keq(0)
    def check_zero(a):
        return a

    assert check_zero(0) == 0
    with pytest.raises(ValueError, match="Value should be equal to 0"):
        check_zero(1)

    @keq(-5)
    def check_negative(a):
        return a

    assert check_negative(-5) == -5
    with pytest.raises(ValueError, match="Value should be equal to -5"):
        check_negative(5)



# External dependencies
import logging
import importlib.util
import mypy.api as mypy_api

# import pysnooper

from flake8.api import legacy as flake8

# Internal dependencies
from kaya_module_sdk.src.exceptions.module_not_found import ModuleNotFoundException
from kaya_module_sdk.src.utils.metadata.equal import EQ
from kaya_module_sdk.src.utils.metadata.greater import GT
from kaya_module_sdk.src.utils.metadata.greater_or_equal import GTE
from kaya_module_sdk.src.utils.metadata.less import LT
from kaya_module_sdk.src.utils.metadata.less_or_equal import LTE
from kaya_module_sdk.src.utils.metadata.max_len import MaxLen
from kaya_module_sdk.src.utils.metadata.maximum import Max
from kaya_module_sdk.src.utils.metadata.min_len import MinLen
from kaya_module_sdk.src.utils.metadata.minimum import Min
from kaya_module_sdk.src.utils.metadata.value_range import ValueRange
from kaya_module_sdk.src.utils.metadata.eq_len import EQLen
from kaya_module_sdk.src.utils.metadata.const import Const
from kaya_module_sdk.src.utils.metadata.not_const import NotConst
from kaya_module_sdk.src.utils.metadata.order import Order

log = logging.getLogger(__name__)


class KVLE:
    """
    [ KVL(E) ]: Kaya Validation Framework (Executor)

    Responsibilities:
      - Run linters (flake8), type checkers (mypy) and module validations.
      - Load and apply metadata constraint rules.
      - Aggregate validation results for rules, metadata, and source.
    """

    def __init__(self, **context):
        self.context = context
        self.check = {
            "rules": {"ok": [], "nok": []},
            "meta": {"ok": [], "nok": []},
            "source": {"ok": [], "nok": []},
        }

    # @pysnooper.snoop()
    def check_package_installed(self, test: dict) -> bool:
        try:
            importlib.util.find_spec(test["package"])
        except ModuleNotFoundError as e:
            raise ModuleNotFoundException(f'Package {test["package"]} is not installed!') from e
        return True

    # @pysnooper.snoop()
    def load_constraint_rules(self, *rules: str) -> dict:
        matches = []
        metadata_classes = [
            EQ,
            GT,
            GTE,
            LT,
            LTE,
            MaxLen,
            Max,
            MinLen,
            Min,
            ValueRange,
            EQLen,
            Const,
            NotConst,
            Order,
        ]
        for rule in rules:
            for mcls in metadata_classes:
                try:
                    instance = mcls(None, None) if ";" in rule else mcls(None)
                    instance.load(rule)
                except Exception:
                    continue
                matches.append(instance)
        # NOTE: Only include those with complete data.
        formatted = {item.__class__.__name__: item for item in matches if None not in item._data.values()}
        return formatted if rules and formatted else {}

    # @pysnooper.snoop()
    def check_rules(self, **kwargs) -> dict:
        """Runs constraint rules checks on module inputs/outputs and aggregates results."""
        package_instance = self.context.get("module_package", {}).get("instance")
        package_name = package_instance.config.name
        submodules = package_instance.modules
        rules_report: dict = {"ok": [], "nok": []}
        for module_name, module_obj in submodules.items():
            error_flag = False
            module_record = {
                "package": package_name,
                "module": module_name,
                "functions": {"main": []},
            }
            # NOTE: Process manifest validations
            manifest = module_obj.manifest
            if not manifest:
                error_flag = True
                module_record["manifest"] = {"required": True, "set": False}
                module_record["error"] = error_flag
                rules_report["nok"].append(module_record)
                continue
            if not isinstance(manifest, dict) or not manifest.get("inputs"):
                error_flag = True
                module_record["manifest"] = {
                    "required": True,
                    "set": True,
                    "valid": False,
                    "value": manifest,
                }
                module_record["error"] = error_flag
                rules_report["nok"].append(module_record)
                continue
            # NOTE: Process input validations.
            for arg in manifest["inputs"]:
                if not arg.get("validations"):
                    continue
                loaded = self.load_constraint_rules(*arg["validations"])
                if arg["validations"] and not loaded:
                    error_flag = True
                for cname in loaded:
                    values = list(loaded[cname]._data.values())
                    if len(values) == 1:
                        values = values[0]
                    module_record["functions"]["main"].append(
                        {
                            "name": cname,
                            "target": "inputs",
                            "verb": cname.lower(),
                            "field": arg["label"],
                            "rule": [cname, values],
                            "error": error_flag,
                        }
                    )
            # NOTE: Process output validations.
            for ret in manifest["outputs"]:
                if not ret.get("validations"):
                    continue
                loaded = self.load_constraint_rules(*ret["validations"])
                if ret["validations"] and not loaded:
                    error_flag = True
                for cname in loaded:
                    values = list(loaded[cname]._data.values())
                    if len(values) == 1:
                        values = values[0]
                    module_record["functions"]["main"].append(
                        {
                            "name": cname,
                            "target": "outputs",
                            "verb": cname.lower(),
                            "field": ret["label"],
                            "rule": [cname, values],
                            "error": error_flag,
                        }
                    )
            module_record["error"] = error_flag
            if error_flag:
                rules_report["nok"].append(module_record)
            else:
                rules_report["ok"].append(module_record)
        self.check["rules"] = rules_report
        return self.check["rules"]

    # @pysnooper.snoop()
    def check_meta(self, module_data: dict, report: bool = True, **kwargs) -> dict:
        """Runs metadata validations and aggregates results."""
        package_instance = self.context.get("module_package", {}).get("instance")
        package_name = package_instance.config.name
        package_version = package_instance.config.version
        submodules = package_instance.modules
        meta_report: dict = {"ok": [], "nok": []}
        for module_name, module_obj in submodules.items():
            error_flag = False
            module_record = {
                "package": package_name,
                "package_version": package_version,
                "module": module_name,
            }
            if not module_obj.config:
                error_flag = True
                module_record["config"] = {"required": True, "set": False}
            for key in module_obj.config._mandatory:
                value = module_obj.config.__dict__.get(key)
                if not value:
                    error_flag = True
                    module_record[key] = {"required": True, "set": False}
                    continue
                valid = (type(value) in (str,)) if key != "version" else (type(value) in (str, float))
                module_record[key] = {
                    "required": True,
                    "set": True,
                    "valid": valid,
                    "value": value,
                }
            manifest = module_obj.manifest
            if not manifest:
                error_flag = True
                module_record["manifest"] = {
                    "required": True,
                    "set": False,
                    "valid": False,
                    "value": manifest,
                }
            elif not isinstance(manifest, dict):
                error_flag = True
                module_record["manifest"] = {
                    "required": True,
                    "set": True,
                    "valid": False,
                    "value": manifest,
                }
            if error_flag:
                meta_report["nok"].append(module_record)
            else:
                meta_report["ok"].append(module_record)
        self.check["meta"] = meta_report
        return self.check["meta"]

    # @pysnooper.snoop()
    def check_source(self, loaded_files: dict, report: bool = True, **kwargs) -> dict:
        """Runs source code validations (flake8, mypy) and aggregates results."""
        source_report: dict = {"ok": [], "nok": []}
        for file_path, file_lines in loaded_files.items():
            errors: list = []
            style_guide = flake8.get_style_guide()
            flake8_report = style_guide.check_files([file_path])
            if flake8_report.total_errors > 0:
                flake8_output = [
                    {
                        error[0]: [error[1], error[2]]
                        for error in flake8_report._application.file_checker_manager.results
                    }
                ]
                errors.append({"tool": "flake8", "output": flake8_output})
            mypy_result = mypy_api.run([file_path])
            mypy_stdout, mypy_stderr, mypy_exit_status = mypy_result
            if mypy_stderr or mypy_exit_status:
                errors.append(
                    {
                        "tool": "mypy",
                        "output": mypy_stdout + mypy_stderr,
                        "exit": mypy_exit_status,
                    }
                )
            if errors:
                source_report["nok"].append({"path": file_path, "errors": errors, "result": "NOK"})
            else:
                source_report["ok"].append({"path": file_path, "errors": [], "result": "OK"})
        self.check["source"] = source_report
        return self.check["source"]



from kaya_module_sdk.src.exceptions.kunimplemented import KayaUnimplementedException


def test_kaya_unimplemented_exception_default_error_code():
    """Test initialization with default error code."""
    exception = KayaUnimplementedException("Feature not implemented")
    assert exception.message == "Feature not implemented"
    assert exception.error_code == 1


def test_kaya_unimplemented_exception_custom_error_code():
    """Test initialization with a custom error code."""
    exception = KayaUnimplementedException("Feature not implemented", 501)
    assert exception.message == "Feature not implemented"
    assert exception.error_code == 501


def test_kaya_unimplemented_exception_string_representation():
    """Test the string representation of the exception."""
    exception = KayaUnimplementedException("Feature not implemented", 501)
    assert str(exception) == "Error 501: Feature not implemented"



class ModuleNotFoundException(Exception):
    message: str
    error_code: int

    def __init__(self, msg: str, err_code: int = 8):
        super().__init__(msg)
        self.message = msg
        self.error_code = err_code

    def __str__(self):
        return f"Error {self.error_code}: {self.message}"



# External dependencies
import importlib
import pytest

# Internal dependencies
from kaya_module_sdk.src.exceptions.module_not_found import ModuleNotFoundException
from kaya_module_sdk.src.testing.kvl_harness import KVL
from kaya_module_sdk.src.testing.kvl_executer import KVLE


def test_check_package_installed_success(monkeypatch):
    # NOTE: Simulate a successful find_spec call.
    def fake_find_spec(package):
        return object()

    import importlib.util

    monkeypatch.setattr(importlib.util, "find_spec", fake_find_spec)
    kvle = KVLE()
    assert kvle.check_package_installed({"package": "dummy_pkg"}) is True


def test_check_package_installed_failure(monkeypatch):
    # NOTE: Simulate failure by raising ModuleNotFoundError.
    def fake_find_spec(package):
        raise ModuleNotFoundError("not found")

    import importlib.util

    monkeypatch.setattr(importlib.util, "find_spec", fake_find_spec)
    kvle = KVLE()
    with pytest.raises(ModuleNotFoundException):
        kvle.check_package_installed({"package": "dummy_pkg"})


def test_load_constraint_rules(monkeypatch, patch_meta_kvle):
    patch_meta_kvle(monkeypatch)
    kvle = KVLE()
    # NOTE: Call load_constraint_rules with a test rule.
    result = kvle.load_constraint_rules("rule1")
    # NOTE: Expect that DummyConstraint was used to load the rule.
    assert "DummyConstraint" in result
    assert result["DummyConstraint"]._data["dummy"] == "value"


def test_check_rules(monkeypatch, mock_module_instance, dummy_constraint):
    context = {"module_package": {"instance": mock_module_instance}}
    kvle = KVLE(**context)

    # NOTE: Patch load_constraint_rules to simulate a successful constraint load.
    def fake_load_constraint_rules(*rules):
        if "rule1" in rules:
            dummy = dummy_constraint(None)
            return {"DummyConstraint": dummy}
        return {}

    monkeypatch.setattr(kvle, "load_constraint_rules", fake_load_constraint_rules)
    result = kvle.check_rules()
    # NOTE: Expect a dict with keys "ok" and "nok"
    assert "ok" in result
    assert "nok" in result
    # NOTE: Since mod2 has manifest=None, it should appear in "nok".
    nok = result["nok"]
    assert any(mod["module"] == "mod2" for mod in nok)


def test_check_meta(monkeypatch, mock_module_instance):
    context = {"module_package": {"instance": mock_module_instance}}
    kvle = KVLE(**context)
    result = kvle.check_meta(module_data={})
    assert "ok" in result
    assert "nok" in result
    nok = result["nok"]
    ok = result["ok"]
    # NOTE: mod2 (missing manifest) should be in nok.
    assert any(mod["module"] == "mod2" for mod in nok)
    # NOTE: mod1 should be valid and appear in ok.
    assert any(mod["module"] == "mod1" for mod in ok)


def test_check_source_success(tmp_path, monkeypatch, dummy_kvle, dummy_kvlr):
    # Create a temporary Python file.
    file_path = tmp_path / "test_source.py"
    content = "print('Hello')"
    file_path.write_text(content)
    # NOTE: Loaded files dictionary (expected format by check_source).
    #   loaded_files = {str(file_path): content.splitlines(keepends=True)}
    kvl_instance = KVL(str(file_path))
    # NOTE: Patch the method search_python_files_in_dir on the KVL instance
    monkeypatch.setattr(kvl_instance, "search_python_files_in_dir", lambda x: [str(file_path)])
    # NOTE: Let load_python_files work normally
    #   Patch get_executer and get_reporter to return dummy objects.
    monkeypatch.setattr(kvl_instance, "get_executer", lambda **kwargs: dummy_kvle())
    monkeypatch.setattr(kvl_instance, "get_reporter", lambda **kwargs: dummy_kvlr())
    report = kvl_instance.check_source(file_path=str(file_path), dump_report=False)
    # NOTE: Dummy reporter wraps the data under key "report"
    assert "report" in report


def test_check_source_errors(tmp_path, monkeypatch, dummy_kvlr_nok, dummy_kvle_source_nok):
    """
    Test that check_source returns a report with errors when simulated errors occur.
    We simulate errors by patching get_executer to return a dummy error result and
    by patching flake8 and mypy to simulate failures.
    """
    file_path = tmp_path / "test_source.py"
    content = "print('Hello')"
    file_path.write_text(content)
    kvl_instance = KVL(str(file_path))
    # NOTE: Patch search_python_files_in_dir on the KVL instance.
    monkeypatch.setattr(kvl_instance, "search_python_files_in_dir", lambda target: [str(file_path)])
    # NOTE: Patch get_executer and get_reporter to return dummy objects that simulate errors.
    monkeypatch.setattr(kvl_instance, "get_executer", lambda **kwargs: dummy_kvle_source_nok())
    monkeypatch.setattr(kvl_instance, "get_reporter", lambda **kwargs: dummy_kvlr_nok())

    # NOTE: Patch flake8 to simulate errors.
    class DummyFlake8Report:
        total_errors = 5
        _application = type(
            "DummyApp",
            (),
            {"file_checker_manager": type("DummyFcm", (), {"results": [("file", 1, 2)]})},
        )()

    dummy_style_guide = type("DummyStyleGuide", (), {"check_files": lambda self, files: DummyFlake8Report()})()
    monkeypatch.setattr("flake8.api.legacy.get_style_guide", lambda: dummy_style_guide())
    # NOTE: Patch mypy to simulate errors.
    monkeypatch.setattr("mypy.api.run", lambda files: ("mypy error", "mypy stderr", 1))
    report = kvl_instance.check_source(file_path=str(file_path), dump_report=False)
    # NOTE: The DummyKVLRForErrors wraps the KVLE result in a dict under "report".
    rep = report.get("report")
    assert rep is not None
    # NOTE: Dummy executer returns a dict with a non-empty "nok" list.
    #   We assume the structure is: { "source": { "ok": [...], "nok": [...] } }
    if "source" in rep:
        nok_list = rep["source"].get("nok", [])
    else:
        nok_list = rep.get("nok", [])
    assert len(nok_list) > 0


def test_check_combined(monkeypatch, import_module):
    monkeypatch.setattr(importlib, "import_module", import_module)
    kvl_instance = KVL(module_name="dummy_module")
    # NOTE: Patch individual check methods to return dummy reports.
    monkeypatch.setattr(kvl_instance, "check_meta", lambda **kwargs: {"meta": "ok"})
    monkeypatch.setattr(kvl_instance, "check_source", lambda **kwargs: {"source": "ok"})
    monkeypatch.setattr(kvl_instance, "check_rules", lambda **kwargs: {"rules": "ok"})
    # NOTE: Call the combined check method explicitly
    combined = kvl_instance.check("meta", "source", "rules", module="dummy_module")
    # NOTE: Expect a dictionary with keys for each target.
    assert "meta" in combined
    assert "source" in combined
    assert "rules" in combined
    # NOTE: Call the combined check method implicitly
    combined = kvl_instance.check("all", module="dummy_module")
    assert "meta" in combined
    assert "source" in combined
    assert "rules" in combined


# CODE DUMP



from kaya_module_sdk.src.utils.metadata.less_or_equal import LTE
from kaya_module_sdk.src.utils.metadata.abstract_metadata import KMetadata
from kaya_module_sdk.src.utils.metadata.abstract_validation import KValidation


def test_lte_initialization_with_float():
    """Test that LTE initializes correctly with a float value."""
    instance = LTE(3.14)
    assert instance._data["less_than_or_equal_to"] == 3.14


def test_lte_initialization_with_str():
    """Test that LTE initializes correctly with a string value."""
    instance = LTE("example")
    assert instance._data["less_than_or_equal_to"] == "example"


def test_lte_str_with_float():
    """Test that __str__ returns the expected string for a float value."""
    instance = LTE(2.71)
    expected = "<=:2.71"
    assert str(instance) == expected


def test_lte_str_with_str():
    """Test that __str__ returns the expected string for a string value."""
    instance = LTE("hello")
    expected = "<=:hello"
    assert str(instance) == expected


def test_lte_load_valid_float():
    """
    Test that load() correctly processes a valid string representation
    where the value can be converted to a float.
    """
    instance = LTE(0)
    result = instance.load("<=:3.14")
    expected = {"less_than_or_equal_to": 3.14}
    assert result == expected
    assert instance._data["less_than_or_equal_to"] == 3.14


def test_lte_load_valid_non_numeric():
    """
    Test that load() correctly processes a valid string representation
    where the value is non-numeric.
    """
    instance = LTE(0)
    result = instance.load("<=:not_a_number")
    expected = {"less_than_or_equal_to": "not_a_number"}
    assert result == expected
    assert instance._data["less_than_or_equal_to"] == "not_a_number"


def test_lte_load_invalid_no_colon():
    """
    Test that load() returns an empty dictionary when the colon is missing.
    """
    instance = LTE(0)
    result = instance.load("<=3.14")
    assert result == {}
    assert instance._data["less_than_or_equal_to"] == 0


def test_lte_load_invalid_wrong_prefix():
    """
    Test that load() returns an empty dictionary when the prefix is incorrect.
    """
    instance = LTE(0)
    result = instance.load("wrong:3.14")
    assert result == {}
    assert instance._data["less_than_or_equal_to"] == 0


def test_lte_load_invalid_extra_segments():
    """
    Test that load() returns an empty dictionary when there are extra segments.
    """
    instance = LTE(0)
    result = instance.load("<=:3.14:extra")
    assert result == {}
    assert instance._data["less_than_or_equal_to"] == 0


def test_lte_inheritance():
    """
    Test that LTE is an instance of both KMetadata and KValidation.
    """
    instance = LTE(3.14)
    assert isinstance(instance, KMetadata)
    assert isinstance(instance, KValidation)



import logging

# #import pysnooper

from subprocess import PIPE, Popen

log = logging.getLogger("KayaPythonModuleSDK")


# #@pysnooper.snoop()
def shell_cmd(command: list, user: str | None = None):
    if not command:
        raise ValueError("No command specified!")
    log.debug("Issuing system command: (%s)", command)
    fmt_command = " ".join(command)
    if user:
        fmt_command = f"su {user} -c '{fmt_command}'"
    with Popen(fmt_command, shell=True, stdout=PIPE, stderr=PIPE) as process:
        output, errors = process.communicate()
        log.debug("Output: (%s), Errors: (%s)", output, errors)
        return (str(output).rstrip("\n"), str(errors).rstrip("\n"), process.returncode)



import pytest

from kaya_module_sdk.src.module.arguments import Args
from kaya_module_sdk.sdk import kaya_io

# TESTS


@pytest.mark.parametrize(
    "kaya_io_decorated_args_type",
    [{"_field1": 10, "_field2": "test"}],
    indirect=True,  # NOTE: Pass parameter to the fixture
)
def test_dynamic_init(kaya_io_decorated_args_type):
    obj = kaya_io_decorated_args_type
    assert isinstance(obj, Args)


@pytest.mark.parametrize(
    "kaya_io_decorated_args_type",
    [{"_field1": 10, "_field2": "test"}],
    indirect=True,
)
def test_setter(kaya_io_decorated_args_type):
    obj = kaya_io_decorated_args_type
    obj.set_field1(20)
    obj.set_field2("updated")
    assert obj.field1 == 20
    assert obj.field2 == "updated"


@pytest.mark.parametrize(
    "kaya_io_decorated_args_type",
    [{"_field1": 10, "_field2": "test"}],
    indirect=True,
)
def test_getter(kaya_io_decorated_args_type):
    obj = kaya_io_decorated_args_type
    assert obj.field1 == 10
    assert obj.field2 == "test"


def test_missing_annotated_field():
    @kaya_io()
    class MissingAnnotatedFieldClass:
        field1: int  # Not Annotated, should be ignored

    obj = MissingAnnotatedFieldClass()
    assert not hasattr(obj, "field1")


@pytest.mark.parametrize(
    "kaya_io_decorated_args_type",
    [{"_field1": None, "_field2": None}],
    indirect=True,
)
def test_init_optional_fields(kaya_io_decorated_args_type):
    obj = kaya_io_decorated_args_type
    assert obj.field1 is None
    assert obj.field2 is None


@pytest.mark.parametrize(
    "kaya_io_decorated_args_type",
    [{"_field1": 10, "_field2": "test"}],
    indirect=True,
)
def test_generated_methods(kaya_io_decorated_args_type):
    obj = kaya_io_decorated_args_type
    assert getattr(obj, "field1")
    assert callable(getattr(obj, "set_field1"))
    assert getattr(obj, "field2")
    assert callable(getattr(obj, "set_field2"))



# #import pysnooper

from functools import partial
from typing import get_type_hints, get_origin, Annotated, get_args


# #@pysnooper.snoop()
def kaya_io():
    def wrapper(cls):
        annotations = get_type_hints(cls, include_extras=True)
        # NOTE: Build the __init__ signature and method body
        parameter_lines, body_lines = build_class_structure(cls, annotations)
        # NOTE: Generate and attach the dynamic `__init__` method
        attach_init_method(cls, parameter_lines, body_lines)
        return cls

    return wrapper


# #@pysnooper.snoop()
def build_class_structure(cls, annotations):
    """
    Build the list of parameters and the initialization body for the class.
    """
    parameter_lines = []
    body_lines = []
    for field_name, field_type in annotations.items():
        # NOTE: Process fields with Annotated types only
        if get_origin(field_type) is not Annotated:
            continue
        if field_name in ["_errors", "_results"]:
            continue
        base_type = get_args(field_type)[0]
        add_getter_and_setter(cls, field_name, base_type)
        #       optional_type = Union[base_type, None]  # Optional[type]
        parameter_lines.append(build_parameter(field_name, base_type))
        body_lines.append(build_body_line(field_name))
    return parameter_lines, body_lines


# #@pysnooper.snoop()
def add_getter_and_setter(cls, field_name, base_type):
    """
    Dynamically create and attach getter and setter methods for a given field.
    """
    getter_func = partial(create_getter, field_name=field_name)
    setter_func = partial(create_setter, field_name=field_name)
    setattr(cls, field_name.lstrip("_"), getter_func())
    setattr(cls, f'set_{field_name.lstrip("_")}', setter_func())


# #@pysnooper.snoop()
def build_parameter(field_name, base_type):
    """
    Build a string representation of a parameter for the dynamic __init__ method.
    """
    return f"{field_name.strip('_')}: {base_type.__name__} | None = None"


# #@pysnooper.snoop()
def build_body_line(field_name):
    """
    Build a line of code for the __init__ method body to set instance attributes.
    """
    return f"if {field_name.strip('_')} is not None: self.set_{field_name.strip('_')}({field_name.strip('_')})"


# #@pysnooper.snoop()
def attach_init_method(cls, parameter_lines, body_lines):
    """
    Dynamically create and attach the __init__ method to the class.
    """
    init_code = f"""def __init__(self, {', '.join(parameter_lines)}):\n\
        super(type(self), self).__init__()\n\
        {'\n        '.join(body_lines)}"""
    namespace = {}
    exec(init_code, globals(), namespace)
    setattr(cls, "__init__", namespace["__init__"])


# NOTE: Placeholder functions for getter and setter creation
# #@pysnooper.snoop()
def create_getter(field_name):
    @property
    def getter(self):
        return getattr(self, field_name)

    return getter


# #@pysnooper.snoop()
def create_setter(field_name):
    def setter(self, value):
        setattr(self, field_name, value)

    return setter



#!/usr/bin/env python

import glob
import os
import pkgutil
import subprocess
import sys

import mypy.api
from pylint.lint import Run
from setuptools import Command, setup, find_packages

# [ NOTE ]: Directory where this file is located
root_dir = os.path.abspath(os.path.dirname(__file__))
PACKAGE_NAME = "kaya_module_sdk"
MYPY_CONFIG_FILE = "conf/setup.mypy.conf"
PYLINT_CONFIG_FILE = "conf/setup.pylint.conf"
FLAKE8_CONFIG_FILE = "conf/setup.flake8.conf"
# Update the version here, do not remove the variable. It's used in CI
SDK_VERSION = "2.1.6"


class MyPyCommand(Command):
    """[ NOTE ]: Custom command to run MyPy checks using `python setup.py
    mypy`."""

    description = "Run MyPy type checks"
    user_options: list[str] = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        user_options = [
            "--disallow-untyped-defs",
            "--config-file",
            MYPY_CONFIG_FILE,
        ]
        user_options.extend(glob.glob("*.py"))
        result = mypy.api.run(user_options)
        print(f"[ MyPy ]: {result}")

        if result[0]:
            print("[ ERROR ]: MyPy found type errors:")
            print(result[0])
            raise SystemExit(1)

        if result[2]:
            print("[ WARNING ]: MyPy found warnings:")
            print(result[2])


class PyLintCommand(Command):
    """[ NOTE ]: Custom command to run PyLint checks using `python setup.py
    pylint`."""

    description = "Run PyLint type checks"
    user_options: list[str] = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        user_options = [
            f"--rcfile={PYLINT_CONFIG_FILE}",
            "--fail-under=8",
            PACKAGE_NAME,
        ]
        Run(user_options)


class Flake8Command(Command):
    description = "Run flake8 checks"
    user_options: list[str] = []

    def initialize_options(self):
        pass

    def finalize_options(self):
        pass

    def run(self):
        user_options = ["flake8", "--config", FLAKE8_CONFIG_FILE, "."]
        subprocess.call(user_options)


def add_package_dir_to_path():
    """[ NOTE ]: Add the package directory to the Python path."""
    package_dir = os.path.join(root_dir, PACKAGE_NAME)
    sys.path.insert(0, package_dir)
    for loader, name, is_pkg in pkgutil.walk_packages([package_dir]):
        # full_name = PACKAGE_NAME + "." + name
        if is_pkg:
            path = loader.path
        else:
            path = os.path.dirname(loader.path)
        sys.path.insert(0, path)


# MISCELLANEOUS

add_package_dir_to_path()
setup_info = dict(
    name="kaya_module_sdk",
    version=SDK_VERSION,
    author="WanoLabs",
    author_email="contact@wanolabs.com",
    url="https://wanolabs.com",
    description="Kaya Module SDK.",
    long_description=open("README.md", "r").read(),
    long_description_content_type="text/markdown",
    # Package info
    packages=find_packages(exclude=["kaya_module_sdk/tst"]),
    include_package_data=True,
    py_modules=["kaya_module_sdk"],
    zip_safe=True,
    test_suite="KAT",
    cmdclass={
        "flake8": Flake8Command,
        "mypy": MyPyCommand,
        "pylint": PyLintCommand,
    },
    install_requires=["pandas"],
    setup_requires=["flake8", "mypy", "pylint", "setuptools"],
    python_requires=">=3.12",
    entry_points={
        "console_scripts": [
            "kaya_module_runner=kaya_module_runner.app:init",
        ]
    },
)

setup(**setup_info)

# CODE DUMP



import logging

from abc import ABC
from typing import Annotated, get_type_hints

from kaya_module_sdk.src.utils.metadata.display_description import DisplayDescription
from kaya_module_sdk.src.utils.metadata.display_name import DisplayName

log = logging.getLogger(__name__)


class Args(ABC):
    _errors: Annotated[
        list,
        DisplayName("Errors"),
        DisplayDescription("Collection of things that went very, very wrong."),
    ]

    _live: Annotated[
        bool,
        DisplayName("Live Execution"),
        DisplayDescription(
            "Input provided by the Systemic runtime. Indicates if the current request is a historical"
            "backfilling request, or a LIVE execution request. Live requests should only return the"
            "last computed datapoint"
        ),
    ]

    def __init__(self):
        self._errors = []
        self._live = True

    @property
    def errors(self) -> list[Exception]:
        return self._errors

    @property
    def live(self) -> bool:
        return self._live

    def set_errors(self, *values: Exception) -> None:
        self._errors += list(values)

    def set_live(self, value: bool) -> None:
        self._live = value

    def metadata(self):
        return get_type_hints(self, include_extras=True)



from kaya_module_sdk.src.utils.metadata.eq_len import EQLen
from kaya_module_sdk.src.utils.metadata.abstract_metadata import KMetadata
from kaya_module_sdk.src.utils.metadata.abstract_validation import KValidation


def test_eqlen_initialization_with_int():
    """Test that EQLen initializes correctly with an integer."""
    instance = EQLen(10)
    assert instance._data["eq_len"] == 10


def test_eqlen_initialization_with_str():
    """Test that EQLen initializes correctly with a string."""
    instance = EQLen("sample")
    assert instance._data["eq_len"] == "sample"


def test_eqlen_str_with_int():
    """Test that __str__ returns the expected string when initialized with an integer."""
    instance = EQLen(15)
    expected = "eqlen:15"
    assert str(instance) == expected


def test_eqlen_str_with_str():
    """Test that __str__ returns the expected string when initialized with a string."""
    instance = EQLen("hello")
    expected = "eqlen:hello"
    assert str(instance) == expected


def test_eqlen_load_valid_int():
    """
    Test that load() correctly processes a valid string representation
    where the value can be converted to an integer.
    """
    instance = EQLen(0)  # initial value is 0
    result = instance.load("eqlen:20")
    expected = {"eq_len": 20}
    assert result == expected
    assert instance._data["eq_len"] == 20


def test_eqlen_load_valid_non_int():
    """
    Test that load() correctly processes a valid string representation
    where the value is non-numeric.
    """
    instance = EQLen(0)
    result = instance.load("eqlen:abc")
    expected = {"eq_len": "abc"}
    assert result == expected
    assert instance._data["eq_len"] == "abc"


def test_eqlen_load_invalid_format_no_colon():
    """
    Test that load() returns an empty dictionary when the colon is missing.
    """
    instance = EQLen(0)
    result = instance.load("eqlenabc")
    assert result == {}
    # The internal data should remain unchanged.
    assert instance._data["eq_len"] == 0


def test_eqlen_load_invalid_prefix():
    """
    Test that load() returns an empty dictionary when the prefix is incorrect.
    """
    instance = EQLen(0)
    result = instance.load("wrong:20")
    assert result == {}
    assert instance._data["eq_len"] == 0


def test_eqlen_load_with_extra_segments():
    """
    Test that load() returns an empty dictionary when there are extra segments.
    """
    instance = EQLen(0)
    result = instance.load("eqlen:20:extra")
    assert result == {}
    assert instance._data["eq_len"] == 0


def test_eqlen_inheritance():
    """
    Test that EQLen is an instance of both KMetadata and KValidation.
    """
    instance = EQLen(0)
    assert isinstance(instance, KMetadata)
    assert isinstance(instance, KValidation)



import pytest

from kaya_module_sdk.src.utils.constraints.minimum import kmin


def test_all_args_above_min():
    """All arguments are above the minimum value"""

    @kmin(5)
    def func(*args):
        return sum(args)

    result = func(5, 6, 7)  # All values are ≥ 5
    assert result == 18


def test_args_equal_to_min():
    """Arguments equal to the minimum value"""

    @kmin(5)
    def func(*args):
        return sum(args)

    result = func(5, 5, 5)  # All values are exactly 5
    assert result == 15


def test_arg_below_min():
    """Argument below the minimum value"""

    @kmin(5)
    def func(*args):
        return sum(args)

    with pytest.raises(ValueError, match="Value should be greater than or equal to 5"):
        func(5, 4)  # 4 is below the minimum value of 5


def test_mixed_arguments():
    """Mixed valid and invalid arguments"""

    @kmin(5)
    def func(*args):
        return sum(args)

    with pytest.raises(ValueError, match="Value should be greater than or equal to 5"):
        func(5, 6, 4)  # 4 is below the minimum


def test_non_numeric_arguments():
    """Non-numeric arguments"""

    @kmin(5)
    def func(*args):
        return " ".join(map(str, args))

    result = func("hello", "world")  # Non-numeric arguments, should not raise an error
    assert result == "hello world"


def test_empty_arguments():
    """Empty arguments"""

    @kmin(5)
    def func(*args):
        return len(args)

    result = func()  # No arguments provided, should not raise an error
    assert result == 0


def test_keyword_arguments_ignored():
    """Keyword arguments are ignored"""

    @kmin(5)
    def func(*args, **kwargs):
        return len(args), kwargs

    result = func(5, 6, key="value")  # Only positional args are checked
    assert result == (2, {"key": "value"})


def test_single_argument_below_min():
    """Single argument below the minimum value"""

    @kmin(5)
    def func(arg):
        return arg

    with pytest.raises(ValueError, match="Value should be greater than or equal to 5"):
        func(4)  # Argument 4 is below the minimum


def test_single_argument_equal_to_min():
    """Single argument equal to the minimum value"""

    @kmin(5)
    def func(arg):
        return arg

    result = func(5)  # Argument equals the minimum of 5
    assert result == 5


def test_large_range_of_arguments():
    """Large range of arguments"""

    @kmin(0)
    def func(*args):
        return min(args)

    result = func(*range(0, 100))  # All values are ≥ 0
    assert result == 0
    with pytest.raises(ValueError, match="Value should be greater than or equal to 0"):
        func(*range(-1, 100))  # -1 is below the minimum


def test_function_without_args():
    """Decorator applied to functions with no *args"""

    @kmin(5)
    def func():
        return "No arguments"

    result = func()  # No *args, should work without issues
    assert result == "No arguments"



import importlib
import pytest

from kaya_module_sdk.src.testing.kvl_harness import KVL
from kaya_module_sdk.src.exceptions.kvl_failure import KVLFailureException


def test_search_python_files_in_dir(tmp_path):
    """Test: search_python_files_in_dir finds only .py files."""
    d = tmp_path / "subdir"
    d.mkdir()
    file1 = d / "file1.py"
    file1.write_text("print('Hello')")
    file2 = d / "file2.txt"
    file2.write_text("Not Python")
    kvl_instance = KVL(str(d))
    found_files = kvl_instance.search_python_files_in_dir(str(d))
    assert file1.as_posix() in found_files
    assert file2.as_posix() not in found_files


def test_load_python_files(tmp_path):
    """Test: load_python_files reads file content correctly."""
    file_path = tmp_path / "dummy.py"
    content = "print('Test')\nprint('Line2')"
    file_path.write_text(content)
    kvl_instance = KVL(str(file_path))
    loaded = kvl_instance.load_python_files([str(file_path)])
    expected = content.splitlines(keepends=True)
    assert loaded[str(file_path)] == expected


def test_load_module_package(monkeypatch, import_module):
    """Test: load_module_package loads a module package and instantiates strategy module."""
    # NOTE: Patch import_module in the current module's namespace.
    monkeypatch.setattr(importlib, "import_module", import_module)
    kvl_instance = KVL(module_name="dummy_module")
    pkg = kvl_instance.load_module_package("dummy_module")
    assert pkg["name"] == "dummy_module"
    # NOTE: Verify that the instance has a config with name 'dummy'
    assert hasattr(pkg["instance"], "config")
    assert pkg["instance"].config.name == "dummy"


def test_get_executer_and_reporter(dummy_kvle, dummy_kvlr):
    """Test: get_executer and get_reporter return the stored dummy objects."""
    kvl_instance = KVL()
    # NOTE: Initially, executer and reporter should be None.
    assert kvl_instance.executer is None
    assert kvl_instance.reporter is None
    # NOTE: Set them manually.
    kvl_instance.executer = dummy_kvle(dummy="data")
    kvl_instance.reporter = dummy_kvlr(dummy="data")
    exe = kvl_instance.get_executer()
    rep = kvl_instance.get_reporter()
    assert isinstance(exe, dummy_kvle)
    assert isinstance(rep, dummy_kvlr)


def test_check_rules_no_module():
    """Test: check_rules raises exception if module argument is missing."""
    kvl_instance = KVL()
    with pytest.raises(KVLFailureException):
        kvl_instance.check_rules(module=None)


def test_check_rules_success(monkeypatch, import_module, dummy_kvle, dummy_kvlr):
    """Test: check_rules success path."""
    # NOTE: Patch import_module so that load_module_package does not fail.
    monkeypatch.setattr(importlib, "import_module", import_module)
    # NOTE: Create a KVL instance with module_name.
    kvl_instance = KVL(module_name="dummy_module")
    # NOTE: Patch get_executer and get_reporter to return dummy objects.
    monkeypatch.setattr(kvl_instance, "get_executer", lambda **kwargs: dummy_kvle())
    monkeypatch.setattr(kvl_instance, "get_reporter", lambda **kwargs: dummy_kvlr())
    # NOTE: Call check_rules and expect a report with key "report" from dummy_kvlr.
    report = kvl_instance.check_rules(module="dummy_module", dump_report=False)
    assert "report" in report


def test_check_meta_no_module():
    """Test: check_meta raises exception if module argument is missing."""
    kvl_instance = KVL()
    with pytest.raises(KVLFailureException):
        kvl_instance.check_meta(module=None)


def test_check_meta_success(monkeypatch, import_module, dummy_kvle, dummy_kvlr):
    """Test: check_meta success path."""
    monkeypatch.setattr(importlib, "import_module", import_module)
    kvl_instance = KVL(module_name="dummy_module")
    monkeypatch.setattr(kvl_instance, "get_executer", lambda **kwargs: dummy_kvle())
    monkeypatch.setattr(kvl_instance, "get_reporter", lambda **kwargs: dummy_kvlr())
    report = kvl_instance.check_meta(module="dummy_module", dump_report=False)
    assert "report" in report


def test_check_source_success(tmp_path, monkeypatch, dummy_kvle, dummy_kvlr):
    """Test: check_source success path."""
    # NOTE: Create a temporary Python file.
    file_path = tmp_path / "test_source.py"
    file_path.write_text("print('Hello')")
    kvl_instance = KVL(str(file_path))
    # NOTE: Patch search_python_files_in_dir to return our file.
    monkeypatch.setattr(kvl_instance, "search_python_files_in_dir", lambda x: [str(file_path)])
    # NOTE: Let load_python_files work normally. Patch get_executer and get_reporter.
    monkeypatch.setattr(kvl_instance, "get_executer", lambda **kwargs: dummy_kvle())
    monkeypatch.setattr(kvl_instance, "get_reporter", lambda **kwargs: dummy_kvlr())
    report = kvl_instance.check_source(file_path=str(file_path), dump_report=False)
    assert "report" in report


def test_check_combined(monkeypatch, import_module):
    """Test: Combined check method (check) returns all targets."""
    monkeypatch.setattr(importlib, "import_module", import_module)
    kvl_instance = KVL(module_name="dummy_module")
    # NOTE: Patch individual checks.
    monkeypatch.setattr(kvl_instance, "check_meta", lambda **kwargs: {"meta": "ok"})
    monkeypatch.setattr(kvl_instance, "check_source", lambda **kwargs: {"source": "ok"})
    monkeypatch.setattr(kvl_instance, "check_rules", lambda **kwargs: {"rules": "ok"})
    combined = kvl_instance.check("meta", "source", "rules", module="dummy_module")
    assert "meta" in combined
    assert "source" in combined
    assert "rules" in combined



from kaya_module_sdk.src.exceptions.tests_not_found import (
    TestsNotFoundException as TNFException,
)


def test_tests_not_found_exception_default_error_code():
    """Test initialization with default error code."""
    exception = TNFException("Tests not found")
    assert exception.message == "Tests not found"
    assert exception.error_code == 5


def test_tests_not_found_exception_custom_error_code():
    """Test initialization with a custom error code."""
    exception = TNFException("Tests not found", 404)
    assert exception.message == "Tests not found"
    assert exception.error_code == 404


def test_tests_not_found_exception_string_representation():
    """Test the string representation of the exception."""
    exception = TNFException("Tests not found", 404)
    assert str(exception) == "Error 404: Tests not found"



from kaya_module_sdk.src.utils.metadata.max_len import MaxLen
from kaya_module_sdk.src.utils.metadata.abstract_metadata import KMetadata
from kaya_module_sdk.src.utils.metadata.abstract_validation import KValidation


def test_maxlen_initialization_with_float():
    """Test that MaxLen initializes correctly with a float value."""
    instance = MaxLen(10.5)
    assert instance._data["max_len"] == 10.5


def test_maxlen_initialization_with_str():
    """Test that MaxLen initializes correctly with a string value."""
    instance = MaxLen("abc")
    assert instance._data["max_len"] == "abc"


def test_maxlen_str_with_float():
    """Test that __str__ returns the correct string for a float value."""
    instance = MaxLen(42.0)
    expected = "maxlen:42.0"
    assert str(instance) == expected


def test_maxlen_str_with_str():
    """Test that __str__ returns the correct string for a string value."""
    instance = MaxLen("test")
    expected = "maxlen:test"
    assert str(instance) == expected


def test_maxlen_load_valid_float():
    """
    Test that load() correctly processes a valid string representation
    where the value can be converted to a float.
    """
    instance = MaxLen(0)
    result = instance.load("maxlen:100.5")
    expected = {"max_len": 100.5}
    assert result == expected
    assert instance._data["max_len"] == 100.5


def test_maxlen_load_valid_non_numeric():
    """
    Test that load() correctly processes a valid string representation
    where the value is non-numeric.
    """
    instance = MaxLen(0)
    result = instance.load("maxlen:hello")
    expected = {"max_len": "hello"}
    assert result == expected
    assert instance._data["max_len"] == "hello"


def test_maxlen_load_invalid_no_colon():
    """
    Test that load() returns an empty dict when the colon is missing.
    """
    instance = MaxLen(0)
    result = instance.load("maxlen100")
    assert result == {}
    assert instance._data["max_len"] == 0


def test_maxlen_load_invalid_wrong_prefix():
    """
    Test that load() returns an empty dict when the prefix is incorrect.
    """
    instance = MaxLen(0)
    result = instance.load("wrong:100")
    assert result == {}
    assert instance._data["max_len"] == 0


def test_maxlen_load_invalid_extra_segments():
    """
    Test that load() returns an empty dict when there are extra segments.
    """
    instance = MaxLen(0)
    result = instance.load("maxlen:100:extra")
    assert result == {}
    assert instance._data["max_len"] == 0


def test_maxlen_inheritance():
    """
    Test that MaxLen is an instance of both KMetadata and KValidation.
    """
    instance = MaxLen(10)
    assert isinstance(instance, KMetadata)
    assert isinstance(instance, KValidation)



from kaya_module_sdk.src.exceptions.module_not_found import ModuleNotFoundException


def test_module_not_found_exception_default_error_code():
    """Test initialization with default error code."""
    exception = ModuleNotFoundException("Module not found")
    assert exception.message == "Module not found"
    assert exception.error_code == 8


def test_module_not_found_exception_custom_error_code():
    """Test initialization with a custom error code."""
    exception = ModuleNotFoundException("Module not found", 404)
    assert exception.message == "Module not found"
    assert exception.error_code == 404


def test_module_not_found_exception_string_representation():
    """Test the string representation of the exception."""
    exception = ModuleNotFoundException("Module not found", 404)
    assert str(exception) == "Error 404: Module not found"



import pytest

from kaya_module_sdk.src.utils.metadata.display_description import DisplayDescription
from kaya_module_sdk.src.utils.metadata.display_name import DisplayName
from kaya_module_sdk.src.utils.metadata.min_len import MinLen

# from kaya_module_sdk.src.utils.constraints.min_len import kminlen
from kaya_module_sdk.src.module.returns import Rets


def test_initialization():
    """Test initialization and property access"""
    r = Rets()
    # NOTE: Check that the _results and _errors lists are empty on initialization
    assert r.results == [], "Expected _results to be an empty list upon initialization."
    assert r.errors == [], "Expected _errors to be an empty list upon initialization."


def test_set_results():
    """Test the 'set_results' method (to add values to _results)"""
    r = Rets()
    # NOTE: Add some results
    r.set_results(1, 2, 3)
    # NOTE: Check if the values are added correctly
    assert r.results == [1, 2, 3], "Expected _results to contain [1, 2, 3]."


def test_set_errors():
    """Test the 'set_errors' method (to add values to _errors)"""
    r = Rets()
    # NOTE: Add some errors
    r.set_errors(Exception("Error 1"), Exception("Error 2"))
    # NOTE: Check if the errors are added correctly
    assert len(r.errors) == 2, "Expected _errors to contain 2 exceptions."
    assert str(r.errors[0]) == "Error 1", "Expected the first error to be 'Error 1'."
    assert str(r.errors[1]) == "Error 2", "Expected the second error to be 'Error 2'."


def test_property_getters():
    """Test that the `results` and `errors` properties return correct values"""
    r = Rets()
    # NOTE: Set some results and errors
    r.set_results(10, 20, 30)
    r.set_errors(Exception("Test error"))
    # NOTE: Check that the results and errors properties return the correct lists
    assert r.results == [10, 20, 30], "Expected _results to contain [10, 20, 30]."
    assert len(r.errors) == 1, "Expected _errors to contain 1 exception."
    assert str(r.errors[0]) == "Test error", "Expected the first error to be 'Test error'."


def test_metadata():
    r = Rets()
    metadata = r.metadata()
    # NOTE: Check if '_results' and '_errors' are present in the metadata
    assert "_results" in metadata, "Metadata should include '_results'."
    assert "_errors" in metadata, "Metadata should include '_errors'."
    # NOTE: Check if the annotations for '_results' are correctly applied
    results_annotations = metadata["_results"]
    assert hasattr(results_annotations, "__args__"), "Annotations for '_results' should have '__args__'."
    # NOTE: The __args__ in Annotated contains only the type (list[typing.Any]), not the metadata
    # So, we need to inspect the metadata more carefully by checking the annotations directly.
    if len(results_annotations.__args__) == 1:
        # NOTE: The second argument for Annotated should be the metadata
        metadata_items = getattr(results_annotations, "__metadata__", [])
        display_name = None
        display_description = None
        min_len = None
        # NOTE: Iterate through the metadata and find the expected annotations
        for item in metadata_items:
            if isinstance(item, DisplayName):
                display_name = item
            elif isinstance(item, DisplayDescription):
                display_description = item
            elif isinstance(item, MinLen):
                min_len = item
        # NOTE: Verify if the annotations are correctly applied
        assert display_name is not None, "Expected DisplayName annotation for '_results'."
        assert display_name._data.get("name") == "Result", "DisplayName should match the expected value."
        assert display_description is not None, "Expected DisplayDescription annotation for '_results'."
        assert (
            display_description._data.get("description") == "Module computation results."
        ), "DisplayDescription should match the expected value."
        assert min_len is not None, "Expected MinLen annotation for '_results'."
        assert min_len._data.get("min_len") == 1, "MinLen should be 1."
    else:
        pytest.fail("Annotations for '_results' do not contain the expected metadata.")


def test_kminlen_constraint():
    """Test kminlen constraint applied on set_results method"""
    r = Rets()
    # NOTE: Check that we can add at least one result
    r.set_results(1)
    assert r.results == [1], "Expected _results to contain [1]."
    # NOTE: Check that if no results are added, it raises a ValueError due to kminlen
    with pytest.raises(ValueError, match="Composite type length should not be below 1"):
        r.set_results()


def test_results_and_errors():
    """Test if the 'results' and 'errors' properties are being set correctly"""
    r = Rets()
    # NOTE: Directly set the attributes
    r._results = [1, 2, 3]
    r._errors = [Exception("Error 1")]
    # NOTE: Ensure properties return correct values
    assert r.results == [1, 2, 3], "Expected _results to be [1, 2, 3]."
    # NOTE: Check the message of the exception rather than the instance itself
    assert [str(e) for e in r.errors] == ["Error 1"], "Expected _errors to be ['Error 1']."


def test_min_len_constraint():
    """Test if set_results method raises ValueError when there are fewer results than required (due to MinLen)"""
    r = Rets()
    # NOTE: This should raise a ValueError because kminlen is applied and the list is empty
    with pytest.raises(ValueError, match="Composite type length should not be below 1"):
        r.set_results()


# CODE DUMP



from kaya_module_sdk.src.utils.metadata.abstract_metadata import KMetadata
from kaya_module_sdk.src.utils.metadata.abstract_validation import KValidation


class ValueRange(KMetadata, KValidation):
    def __init__(self, min_val: float | str, max_val: float | str) -> None:
        self._data = {
            "min_val": min_val,
            "max_val": max_val,
        }

    def __str__(self) -> str:
        return f'range:{self._data["min_val"]};{self._data["max_val"]}'

    def load(self, str_repr: str) -> dict:
        segmented = str_repr.split(":")
        if not len(segmented) == 2 or segmented[0] != "range":
            return {}
        segmented = segmented[1].split(";")
        if not len(segmented) == 2:
            return {}
        try:
            self._data.update(
                {
                    "min_val": float(segmented[0]),
                    "max_val": float(segmented[1]),
                }
            )
        except Exception:
            self._data.update(
                {
                    "min_val": segmented[0],
                    "max_val": segmented[1],
                }
            )
        return self._data
